/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * <BR><BR>
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * <BR><BR>
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * <BR><BR>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * <BR><BR>
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * <BR><BR>
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */

 




package org.greenstone.gatherer.cdm;
/**************************************************************************************
 * Title:        Gatherer
 * Description:  The Gatherer: a tool for gathering and enriching a digital collection.
 * Copyright:    Copyright (c) 2001
 * Company:      The University of Waikato
 * Written:      03/05/02
 * Revised:      17/11/02
 **************************************************************************************/
import java.util.Collections;
import java.util.Comparator;
import java.util.StringTokenizer;
import java.util.Vector;
import org.greenstone.gatherer.cdm.CollectionMeta;
import org.greenstone.gatherer.msm.MSMUtils;
import org.greenstone.gatherer.util.Utility;
import org.w3c.dom.Element;
/** This class encapsulates a single indexing pair.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.2
 */
public class Index
    implements Comparable {
    /** A refernce to the main manager for access to other sub-managers. */
    private CollectionDesignManager manager = null;
    /** The level of this index. */
    private int level = 0;
    /** The sources for data this index is built apon, which are either fully qualified metadata element names or 'text'. */
    private Vector sources = null;
    /** An element in the index level enumeration. */
    static public final int DOCUMENT = 0;
    /** An element in the index level enumeration. */
    static public final int PARAGRAPH = 1;
    /** An element in the index level enumeration. */
    static public final int SECTION = 2;
    /** An values of items in the index level enumeration. */
    static public final String LEVEL[] = {"document","paragraph","section"};

    /** Constructor.
     * @param level The level of this index as a <strong>String</string>.
     * @param metadata The fully qualified name of the metadata this index is built on as a <strong>String</strong>, or <i>null</i> in which case the data defaults to "text".
     */
    public Index(int level, Vector sources, CollectionDesignManager manager) {
	this.level = level;
	this.manager = manager;
	this.sources = sources;
	if(this.sources == null) {
	    this.sources = new Vector();
	    this.sources.add("text");
	}
    }
    /** Method to compare two indexes.
	  * @param object The other index as an <strong>Object</strong>.
	  * @return An <i>int</i> which indicates how the indexes compare.
	  * @see java.lang.String
	  */
    public int compareTo(Object object) {
	return toString(false).compareTo(object.toString());
    }
    /** Method to test for the equality of two indexes.
	  * @param object The other index as an <strong>Object</strong>.
	  * @return A <i>boolean</i> which is <i>true</i> if the two indexes are equal, <i>false</i> otherwise.
	  */
    public boolean equals(Object object) {
	if(compareTo(object) == 0) {
	    return true;
	}
	return false;
    }
    /** Method to get the data source of this index.
	  * @return A <strong>String</string> which is a comma separated list of either fully qualified names of an assigned metadata elements, or "text".
	  */
    public String getData() {
	String result = "";
	Collections.sort(sources, new IndexComparator());
	for(int i = 0; i < sources.size(); i++) {
	    result = result + sources.get(i).toString();
	    if(i < sources.size() - 1) {
		result = result + ",";
	    }
	}
	return result;
    }
    /** Method to get the data source of this index.
     * @return A <strong>String</string> which is a comma separated list of either fully qualified names of an assigned metadata elements, or "text".
     * Note, greenstone extracted metadata do not have the ex. 
     */
    public String getDataConfig() {
	String result = "";
	Collections.sort(sources, new IndexComparator());
	for(int i = 0; i < sources.size(); i++) {
	    String source = sources.get(i).toString();
	    if (source.startsWith(Utility.EXTRACTED_METADATA_NAMESPACE + MSMUtils.NS_SEP)) {
		// remove the ex. bit
		source = source.substring(source.indexOf(MSMUtils.NS_SEP)+1);
	    }
	    result = result + source;
	    if(i < sources.size() - 1) {
		result = result + ",";
	    }
	}
	return result;
    }
	
    /** Method to get the value of level.
	  * @return The value of level as a <strong>String</strong>.
	  */
    public int getLevel() {
	return level;
    }
    /** Method to get the value of metadata.
	  * @return The value of metadata as an <strong>Vector</strong>.
	  */
    public Vector getMetadata() {
	return sources;
    }
    /** Method to retrieve this indexes name.
	  * @return A <strong>String</strong>.
	  */
    public String getName() {
	if(manager != null) {
	    String name = LEVEL[level] + ":" + getData();
	    Language language = manager.languages.getDefaultLanguage();
	    CollectionMeta metadata = manager.collectionmetadatum.getMetadata(name, language, true);
	    if(metadata != null) {
		return metadata.getValue();
	    }
	}
	return "";
    
    }
    /** Method to turn this object into a string representation ready to be placed in the collection configuration file.
	  * @return A <strong>String</strong> containing the information of this class.
	  */
    public String toString() {
	return LEVEL[level] + ":" + getData();
    }
    /** Retrieve a textual representation of this index.
	* @param show_name <i>true</i> if you want the name of this index, <i>false</i> for the gsdl index reference.
	*/
    public String toString(boolean show_name) {
	if(show_name) {
	    return getName();
	}
	return LEVEL[level] + ":" + getData();
    }
    
    /** Retrieve a textual representation of this index, specifically for the configuration files. Extracted metadata is shown to teh user and stored as ex.MD, however in the config file it should have no namespace */
    public String toStringConfig() {

	return LEVEL[level] + ":" + getDataConfig();
    }

    /** Method to parse a configuration file index specification into an Index
     * object. returns null if invalid syntax */
    public static Index parseIndexConfig(String entry, CollectionDesignManager manager) {
	// all (mg) indexes must start with level:
	if(entry.indexOf(":") == -1) {
	    return null;
	}
	
	String level_str = entry.substring(0, entry.indexOf(":"));
	String sources_raw = entry.substring(entry.indexOf(":") + 1);
	Vector sources = new Vector();
	StringTokenizer st = new StringTokenizer(sources_raw, ",");
	while(st.hasMoreTokens()) {
	    String token = st.nextToken();
	    // We may have to replace old : with whatever namespace separator we are using.
	    token = token.replace(':', MSMUtils.NS_SEP);
	    // if there is no namespace, we need to add one, except for text
	    if (token.indexOf(MSMUtils.NS_SEP)==-1) {
		if (!token.equals("text")) {
		    token = Utility.EXTRACTED_METADATA_NAMESPACE+MSMUtils.NS_SEP+token;
		}
	    }
	    
	    sources.add(token);
	}
	Index index = null;
	for(int i = 0; i < Index.LEVEL.length; i++) {
	    if(level_str.equals(Index.LEVEL[i])) {
		return new Index(i, sources, manager);
		
	    }
	}
	// somethings gone wrong
	return null;
	
    }
    
    /** A custom comparator for comparing Indexes. */
    private class IndexComparator
	implements Comparator {
	/** Method to compare two objects, which may be either indexes or strings.
	 * @param object1 One object as an <strong>Object</strong>.
	 * @param object2 Another object as an <strong>Object</strong>.
	 * @return An <i>int</i> which indicates how they compare.
	 * @see java.lang.String
	 */
	public int compare(Object object1, Object object2) {
	    if(object1 instanceof Index && object2 instanceof Index) {
		Index index1 = (Index)object1;
		Index index2 = (Index)object2;
		return index1.toString(false).compareTo(index2.toString(false));
	    }
	    else if(object1 instanceof Index) {
		Index index = (Index) object1;
		return index.toString(false).compareTo(object2.toString());
	    }
	    else if(object2 instanceof Index) {
		Index index = (Index) object2;
		return object1.toString().compareTo(index.toString(false));
	    }
	    return object1.toString().compareTo(object2.toString());
	}
	/** Method to test for the equality of two objects, which may be indexes or strings.
			* @param object Another object as an <strong>Object</strong>.
			* @return A <i>boolean</i> which is <i>true</i> if the two objects are equal, <i>false</i> otherwise.
			*/
	public boolean equals(Object object) {
	    if(compareTo(object) == 0) {
		return true;
	    }
	    return false;
	}
    }
}
